#include "StdAfx.h"
#include "ETUDriverSinkWrapper.h"

CETUDriverSinkWrapper::CETUDriverSinkWrapper(void)
{
	mbcb = NULL;
	mscb = NULL;
	mfcb = NULL;

	pSink = NULL;
	InitializeSink();
}

CETUDriverSinkWrapper::~CETUDriverSinkWrapper(void)
{
	FinalizeSink();
}

void CETUDriverSinkWrapper::InitializeSink()
{
	if (pSink != NULL)
		return;

	HRESULT hr = CoInitialize(NULL);
	hr = CoCreateInstance(__uuidof(ETUDriverSink), NULL, CLSCTX_INPROC_SERVER, __uuidof(IETUDriverSink), (LPVOID*)&pSink);
	if(pSink)
		pSink->Start();

	pSink->SetGazeProcessor(IGazeProcessorPtr((IGazeProcessor*)this));
}

void CETUDriverSinkWrapper::FinalizeSink()
{
	if (pSink != NULL)
		pSink->Stop();
}

void CETUDriverSinkWrapper::StartTracking()
{
	if (pSink != NULL)
		pSink->StartTracking();
}

void CETUDriverSinkWrapper::StopTracking()
{
	if (pSink != NULL)
		pSink->StopTracking();
}

bool CETUDriverSinkWrapper::IsTracking()
{
	return (pSink == NULL) ? false : (pSink->IsTracking() != S_FALSE);
}
	
void CETUDriverSinkWrapper::RegisterBlinkCallback(blinkCallback callback)
{
	mbcb = callback;
}

void CETUDriverSinkWrapper::RegisterSampleCallback(sampleCallback callback)
{
	mscb = callback;
}

void CETUDriverSinkWrapper::RegisterFixationCallback(fixationCallback callback)
{
	mfcb = callback;
}

void CETUDriverSinkWrapper::UnRegisterBlinkCallback(blinkCallback callback)
{
	mbcb = NULL;
}

void CETUDriverSinkWrapper::UnRegisterSampleCallback(sampleCallback callback)
{
	mscb = NULL;
}

void CETUDriverSinkWrapper::UnRegisterFixationCallback(fixationCallback callback)
{
	mfcb = NULL;
}

void CETUDriverSinkWrapper::ProcessBlink(SiETUDBlink blink)
{
	if (mbcb != NULL){
		(*mbcb)((int)blink.Number, (int)blink.Time, (int)blink.Duration, (int)blink.Eye);
	}
}

void CETUDriverSinkWrapper::ProcessSample(SiETUDSample sample)
{
	if (mscb != NULL)
	{
		(*mscb)((int)sample.Number, (int)sample.Time, (int)sample.Event[0], (int)0, (int)sample.Pupil[0], (int)sample.X[0], (int)sample.Y[0]);
		(*mscb)((int)sample.Number, (int)sample.Time, (int)sample.Event[1], (int)1, (int)sample.Pupil[1], (int)sample.X[1], (int)sample.Y[1]);
	}
}

void CETUDriverSinkWrapper::ProcessFixation(SiETUDFixation fixation)
{
	if (mfcb != NULL){
		(*mfcb)((int)fixation.Number, (int)fixation.Time, (int)fixation.Duration, (int)fixation.Eye, (int)fixation.X, (int)fixation.Y);
	}
}

void CETUDriverSinkWrapper::ShowOptions()
{
	if (pSink != NULL)
		pSink->Options();

	//_beginthreadex(NULL, 0, CETUDriverSinkWrapper::optionsThread, NULL, 0,NULL);
}

void CETUDriverSinkWrapper::ShowSources()
{
	if (pSink != NULL)
		pSink->Source();

	//_beginthreadex(NULL, 0, CETUDriverSinkWrapper::sourcesThread, NULL, 0,NULL);
}

void CETUDriverSinkWrapper::Calibrate()
{
	if (pSink != NULL)
		pSink->Calibrate();

	//_beginthreadex(NULL, 0, CETUDriverSinkWrapper::calibrateThread, NULL, 0,NULL);
}

/*unsigned int __stdcall CETUDriverSinkWrapper::optionsThread(void* args)
{
	if (pSink != NULL)
		pSink->Options();

	return 0;
}

unsigned int __stdcall CETUDriverSinkWrapper::calibrateThread(void* args)
{
	if (pSink != NULL)
		pSink->Calibrate();
	
	return 0;
}

unsigned int __stdcall CETUDriverSinkWrapper::sourcesThread(void* args)
{
	if (pSink != NULL)
		pSink->Source();
	
	return 0;
}*/