package GD.Data;

public abstract class GazeEntryBase {
	/**
	 * Gaze data entry ordinal number. It is counted from the moment of tracking
	 * start
	 */
	int mOrdinal;

	/**
	 * Gaze data entry time stamp. It is an amount of milliseconds passed from
	 * the moment of gaze tracking start
	 */
	int mTimestamp;

	/**
	 * Constructor
	 * 
	 * @param ordinal
	 * @param timestamp
	 */
	public GazeEntryBase(int ordinal, int timestamp) {
		mTimestamp = timestamp;
		mOrdinal = ordinal;
	}

	/**
	 * Compare two gaze entries
	 * 
	 * @param other
	 * @return
	 */
	public boolean Equals(GazeEntryBase other) {
		return other.mOrdinal == mOrdinal;
	}

	/**
	 * Get length of byte representation of the gaze data entry
	 * 
	 * @return Length of the message
	 */
	abstract public int MessageLength();

	/**
	 * Convert gaze data entry into a byte sequence that can be transfered to
	 * the server
	 * 
	 * @return Byte array, prepared for sending via network
	 */
	abstract public byte[] MessageStruct();

	/**
	 * Get type of the entry
	 * 
	 * @return {@link EGDEntryType EGDEntryType} value. Representation type of
	 *         the entry
	 */
	abstract public EEntryType Type();

	/**
	 * Log file string representation
	 * 
	 * @return
	 */
	abstract public String ToLogString();
}
