package GD.Data;

/**
 * Blink gaze data. It is generated by tracker when the user blinks. This gaze
 * data do not contain coordinates, so it cannot be tracked against screen
 * coordinate system. It was added for further possible use in statistical
 * evaluations.
 * 
 * @see IGDEntry
 * @see JGDFixation
 * @see JGDSingleSample
 * 
 * @version 2.1.0
 * @author Andrey Shipilov
 * 
 */
public class GazeEntryBlink extends GazeEntryBase {
    public static final int MSG_SIZE = 16;
    
	/** Duration of blink */
	int mDuration;

	/** Eye which generated blink event */
	EEye mEye;

	/**
	 * Constructor
	 * 
	 * @param ordinal
	 *            Ordinal number of blink
	 * @param time
	 *            Time stamp of blink
	 * @param duration
	 *            Duration of blink
	 * @param eye
	 *            Eye, generated blink event
	 */
	public GazeEntryBlink(int ordinal, int time, int duration, int eye) {
		super(ordinal, time);
		this.mDuration = duration;
		this.mEye = EEye.parse(eye);
	}

	@Override
	public int MessageLength() {
		return MSG_SIZE;
	}

	@Override
	public byte[] MessageStruct() {
		byte[] msg = new byte[MessageLength()];

		msg[0] = (byte) (mOrdinal & 0xff);
                msg[1] = (byte) ((mOrdinal >> 8) & 0xff);
                msg[2] = (byte) ((mOrdinal >> 16) & 0xff);
		msg[3] = (byte) ((mOrdinal >> 24) & 0xff);

                msg[4] = (byte) (mTimestamp & 0xff);
                msg[5] = (byte) ((mTimestamp >> 8) & 0xff);
                msg[6] = (byte) ((mTimestamp >> 16) & 0xff);
		msg[7] = (byte) ((mTimestamp >> 24) & 0xff);

                msg[8] = (byte) (mDuration & 0xff);
                msg[9] = (byte) ((mDuration >> 8) & 0xff);
                msg[10] = (byte) ((mDuration >> 16) & 0xff);
		msg[11] = (byte) ((mDuration >> 24) & 0xff);

		int ieye = EEye.parse(mEye);
                msg[12] = (byte) (ieye & 0xff);
                msg[13] = (byte) ((ieye >> 8) & 0xff);
                msg[14] = 0;
                msg[15] = 0;

		return msg;
	}

	@Override
	public EEntryType Type() {
		return EEntryType.ET_BLINK;
	}

	@Override
	/**
	 * Format :
	 * 		Timestamp	Type	Ordinal	Duration	Eye	PupilSize	Event X Y
	 */
	public String ToLogString() {
            String str = "";
            str += String.valueOf(mTimestamp) + "\t";
            str += String.valueOf(Type()) + "\t";
            str += String.valueOf(mOrdinal) + "\t";
            str += String.valueOf(mDuration) + "\t";
            str += String.valueOf(mEye) + "\t";
            str += String.valueOf("-") + "\t";
            str += String.valueOf("-") + "\t";
            str += String.valueOf("-") + "\t";
            str += String.valueOf("-");
            
            return str;
	}
}
