package GD.Data;

import java.awt.Point;

/**
 * Fixation gaze data entry. It contain data of generated by tracker fixation.
 * It includes coordinates and it can be tracked and drawn on the client screen.
 * 
 * @see IGDEntry
 * @see IGDTrackableEntry
 * @see JGDBlink
 * @see JGDSingleSample
 * 
 * @version 2.0.0
 * @author Andrey Shipilov
 */
public class GazeEntryFixation extends GazeEntryBase implements ICoordinates {
    public static final int MSG_SIZE = 24;
    
	/** Duration of fixation */
	int mDuration;

	/** Eye which generated fixation event */
	EEye mEye;

	/** Coordinates of the gaze. It is stored in server coordinate system */
	Point mCoords;

	/**
	 * Constructor
	 * 
	 * @param ordinal
	 *            Ordinal number of blink
	 * @param time
	 *            Time stamp of blink
	 * @param duration
	 *            Duration of blink
	 * @param eye
	 *            Eye, generated blink event
	 * @param coords
	 *            Coordinates of the gaze
	 */
	public GazeEntryFixation(int ordinal, int time, int duration, int eye,
			Point coords) {
		super(ordinal, time);

		mDuration = duration;
		mEye = EEye.parse(eye);
		mCoords = coords;
	}

	@Override
	public Point Coords() {
		return mCoords;
	}

	@Override
	public int MessageLength() {
		return MSG_SIZE;
	}

	@Override
	public byte[] MessageStruct() {
		byte[] msg = new byte[MessageLength()];

		msg[0] = (byte) (mOrdinal & 0xff);
                msg[1] = (byte) ((mOrdinal >> 8) & 0xff);
                msg[2] = (byte) ((mOrdinal >> 16) & 0xff);
		msg[3] = (byte) ((mOrdinal >> 24) & 0xff);

                msg[4] = (byte) (mTimestamp & 0xff);
                msg[5] = (byte) ((mTimestamp >> 8) & 0xff);
                msg[6] = (byte) ((mTimestamp >> 16) & 0xff);
		msg[7] = (byte) ((mTimestamp >> 24) & 0xff);

                msg[8] = (byte) (mDuration & 0xff);
                msg[9] = (byte) ((mDuration >> 8) & 0xff);
                msg[10] = (byte) ((mDuration >> 16) & 0xff);
		msg[11] = (byte) ((mDuration >> 24) & 0xff);

		int ieye = EEye.parse(mEye);
                msg[12] = (byte) (ieye & 0xff);
                msg[13] = (byte) ((ieye >> 8) & 0xff);
                msg[14] = 0;
                msg[15] = 0;

		msg[16] = (byte) (mCoords.x & 0xff);
                msg[17] = (byte) ((mCoords.x >> 8) & 0xff);
                msg[18] = (byte) ((mCoords.x >> 16) & 0xff);
		msg[19] = (byte) ((mCoords.x >> 24) & 0xff);
		
		msg[20] = (byte) (mCoords.y & 0xff);
		msg[21] = (byte) ((mCoords.y >> 8) & 0xff);
                msg[22] = (byte) ((mCoords.y >> 16) & 0xff);
                msg[23] = (byte) ((mCoords.y >> 24) & 0xff);

		return msg;
	}

	@Override
	public EEntryType Type() {
		return EEntryType.ET_FIXATION;
	}

	@Override
	/**
	 * Format :
	 * 		Timestamp	Type	Ordinal	Duration	Eye	PupilSize	Event X Y
	 */
	public String ToLogString() {
            String str = "";
            str += String.valueOf(mTimestamp) + "\t";
            str += String.valueOf(Type()) + "\t";
            str += String.valueOf(mOrdinal) + "\t";
            str += String.valueOf(mDuration) + "\t";
            str += String.valueOf(mEye) + "\t";
            str += String.valueOf("-") + "\t";
            str += String.valueOf("-") + "\t";
            str += String.valueOf(mCoords.x) + "\t";
            str += String.valueOf(mCoords.y);
            
            return str;
	}
}