package GD.Data;

import java.util.Vector;

/**
 * Class, representing decorator on Vector class. It include a limit for the
 * number of elements. It works in FIFO schema. If number of elements reaches
 * limit then before appending the new element the oldest one will be deleted,
 * so the length of the queue wouldn't exceed limit.
 * 
 * @version 2.0.0
 * @author Andrey Shipilov
 * 
 * @param <E>
 *            Gaze data class
 * 
 * @see JGDBlink
 * @see JGDSingleSample
 * @see JGDFixation
 */
public class Queue<E> {
	/**
	 * Elements storage
	 */
	Vector<E> mEntries;

	/**
	 * Maximum number of elements in a queue
	 */
	int mLimit;

	/**
	 * Constructor
	 * 
	 * @param maxLength
	 *            Maximum number of elements in a queue
	 */
	public Queue(int maxLength) {
		mEntries = new Vector<E>();

		if (maxLength < 1)
			maxLength = 1;

		mLimit = maxLength;
	}

	/**
	 * Add a new element into the queue
	 * 
	 * @param element
	 *            Element to add
	 */
	public void append(E element) {
		while (size() >= mLimit)
			mEntries.removeElementAt(0);

		mEntries.add(element);
	}

	/**
	 * Extract the oldest element of the queue. It is deleted after extracting
	 * 
	 * @return extracted element
	 */
	public E extract() {
		if (size() < 1)
			return null;

		E elem = mEntries.get(0);
		mEntries.removeElementAt(0);

		return elem;
	}

	/**
	 * Get the last element of the queue. It is NOT deleted after getting.
	 * 
	 * @return Element of the queue
	 */
	public E last() {
		if (isEmpty())
			return null;

		return mEntries.get(size() - 1);
	}

	/**
	 * State of the queue
	 * 
	 * @return true if no elements added to the queue
	 */
	public boolean isEmpty() {
		return size() < 1;
	}

	/**
	 * Number of elements in a queue
	 * 
	 * @return number of elements
	 */
	public int size() {
		return mEntries.size();
	}
}
