package GD.Graphics;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.security.spec.MGF1ParameterSpec;
import java.util.HashMap;
import java.util.Iterator;

import GD.Data.Converter;
import GD.Data.EEntryType;
import GD.Data.GazeClient;
import GD.Data.GazeEntryBase;
import GD.Data.ICoordinateConverter;
import GD.Data.ICoordinates;
import GD.Data.IGazeContainer;
import GD.Data.Queue;

public class LastGazeMapProducer extends GazeMapProducer {

    	public LastGazeMapProducer(String markType) throws Exception {
		super(GazeMarkFactory.getMark(markType));
	}

	@Override
	protected Rectangle2D drawShape(Graphics g, IGazeContainer gazes) {
		Rectangle2D upd = new Rectangle();
		try {
                  
			Queue<GazeEntryBase> entries = gazes.Get(mDrawMode);
                        
                        Point clientCoords = null;
                        while(!entries.isEmpty())
                        {
                            GazeEntryBase entry = entries.last();
                            if (!(entry instanceof ICoordinates))
                                return upd;
                            
                            clientCoords = Converter
					.ServerToClient(((ICoordinates) entry).Coords());
                            
                            if (clientCoords.x >= 0 && clientCoords.y >= 0)
                                break;
                        }
                        
                        if (clientCoords != null)
                            upd = mMark.draw(g, clientCoords);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return upd;
	}

	@Override
	protected Rectangle2D drawLabel(Graphics2D g, IGazeContainer gazes,
			String label) {
		Rectangle2D upd = new Rectangle();
		try {
                    
			Queue<GazeEntryBase> entries = gazes.Get(mDrawMode);
			GazeEntryBase entry = entries.last();

			if (!(entry instanceof ICoordinates))
				return upd;

			Point clientCoords = Converter
					.ServerToClient(((ICoordinates) entry).Coords());
			Point anchor = getLabelAnchor(clientCoords);

			g.drawString(label, anchor.x, anchor.y);

			upd = g.getFont().getStringBounds(label, g.getFontRenderContext());
                        upd.setRect(anchor.x, anchor.y, upd.getWidth(), upd.getHeight());

		} catch (Exception e) {
			e.printStackTrace();
		}
		return upd;
	}

	protected Point getLabelAnchor(Point pnt) {
		Rectangle2D size = mMark.getSize();
		Point anchor = new Point((int) (pnt.x + size.getWidth() / 2),
				(int) (pnt.y - size.getHeight() / 2));
		return anchor;
	}

}
