package GazeVnc;

import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Panel;
import java.awt.ScrollPane;
import java.io.EOFException;
import java.io.IOException;
import java.net.ConnectException;
import java.net.NoRouteToHostException;
import java.net.UnknownHostException;

import TightVnc.ButtonPanel;
import TightVnc.VncViewer;
import GD.Data.*;
import GD.GUI.GazeOptionsFrame;
import GD.GUI.IGazeCanvas;
import TightVnc.RfbProto;
import TightVnc.VncCanvas;
import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.util.logging.Level;
import java.util.logging.Logger;

public class GazeVncViewer extends VncViewer {

    /**
     *
     */
    private static final long serialVersionUID = 1L;
    GazeProto mGazeProto;
    GazeButtonPanel mGazeButtonPanel;
    GazeVncCanvas mGazeCanvas;
    GazeOptionsFrame optionsFrame;

    public static void main(String[] argv) {
        


        GazeVncViewer v = new GazeVncViewer();
        v.mainArgs = argv;
        v.inAnApplet = false;
        v.inSeparateFrame = true;

        v.init();
        v.start();
    }

    public void init() {
        super.init();
        vncFrame.setTitle("Gaze TightVNC");

        optionsFrame = new GazeOptionsFrame();
    }

    public void ShowGazeOptions() {
        if (optionsFrame != null) {
            optionsFrame.Show(mGazeCanvas);
        }
    }

    //
    // Read password parameters. If an "ENCPASSWORD" parameter is set,
    // then decrypt the password into the passwordParam string. Otherwise,
    // try to read the "PASSWORD" parameter directly to passwordParam.
    //
    @Override
    public void run() {
        gridbag = new GridBagLayout();
        vncContainer.setLayout(gridbag);

        GridBagConstraints gbc = new GridBagConstraints();
        gbc.gridwidth = GridBagConstraints.REMAINDER;
        gbc.anchor = GridBagConstraints.NORTHWEST;

        if (showControls) {
            buttonPanel = new GazeButtonPanel(this);
            gridbag.setConstraints(buttonPanel, gbc);
            vncContainer.add(buttonPanel);
        }

        try {
            connectAndAuthenticate();
            doProtocolInitialisation();

            // FIXME: Use auto-scaling not only in a separate frame.
            if (options.autoScale && inSeparateFrame) {
                Dimension screenSize;
                try {
                    screenSize = vncContainer.getToolkit().getScreenSize();
                } catch (Exception e) {
                    screenSize = new Dimension(0, 0);
                }
                createCanvas(screenSize.width - 32, screenSize.height - 32);
            } else {
                createCanvas(0, 0);
            }

            gbc.weightx = 1.0;
            gbc.weighty = 1.0;

            if (inSeparateFrame) {

                // Create a panel which itself is resizeable and can hold
                // non-resizeable VncCanvas component at the top left corner.
                Panel canvasPanel = new Panel();
                canvasPanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
                canvasPanel.add(vc);

                // Create a ScrollPane which will hold a panel with VncCanvas
                // inside.
                desktopScrollPane = new ScrollPane(
                        ScrollPane.SCROLLBARS_AS_NEEDED);
                gbc.fill = GridBagConstraints.BOTH;
                gridbag.setConstraints(desktopScrollPane, gbc);
                desktopScrollPane.add(canvasPanel);

                // Finally, add our ScrollPane to the Frame window.
                vncFrame.add(desktopScrollPane);
                vncFrame.setTitle(rfb.desktopName);
                vncFrame.pack();
                vc.resizeDesktopFrame();

            } else {

                // Just add the VncCanvas component to the Applet.
                gridbag.setConstraints(vc, gbc);
                add(vc);
                validate();

            }

            if (showControls) {
                buttonPanel.enableButtons();
            }

            moveFocusToDesktop();
            
            //
            //mGazeCanvas.StartGazeSink();
            //
            
            processNormalProtocol();

        } catch (NoRouteToHostException e) {
            fatalError("Network error: no route to server: " + host, e);
        } catch (UnknownHostException e) {
            fatalError("Network error: server name unknown: " + host, e);
        } catch (ConnectException e) {
            fatalError("Network error: could not connect to server: " + host
                    + ":" + port, e);
        } catch (EOFException e) {
            if (showOfflineDesktop) {
                e.printStackTrace();
                System.out.println("Network error: remote side closed connection");
                if (vc != null) {
                    vc.enableInput(false);
                }
                if (inSeparateFrame) {
                    vncFrame.setTitle(rfb.desktopName + " [disconnected]");
                }
                if (rfb != null && !rfb.closed()) {
                    rfb.close();
                }
                if (showControls && buttonPanel != null) {
                    buttonPanel.disableButtonsOnDisconnect();
                    if (inSeparateFrame) {
                        vncFrame.pack();
                    } else {
                        validate();
                    }
                }
            } else {
                fatalError("Network error: remote side closed connection", e);
            }
        } catch (IOException e) {
            String str = e.getMessage();
            if (str != null && str.length() != 0) {
                fatalError("Network Error: " + str, e);
            } else {
                fatalError(e.toString(), e);
            }
        } catch (Exception e) {
            String str = e.getMessage();
            if (str != null && str.length() != 0) {
                fatalError("Error: " + str, e);
            } else {
                fatalError(e.toString(), e);
            }
        }

    }

    @Override
    protected void readParameters() {
        double scale = readIntParameter("SCALE", 100);

        //@TODO: read name\description
        String name = readParameter("NAME", false);
        String description = readParameter("DESCRIPTION", false);
        //@TODO mGDCanvas.setScalingFactor(scale, scale);

        super.readParameters();
    }

    protected void moveFocusToDesktop() {
        super.moveFocusToDesktop();
    }

    protected void connectAndAuthenticate() throws Exception {
        showConnectionStatus("Initializing...");
		if (inSeparateFrame) {
			vncFrame.pack();
			vncFrame.show();
		} else {
			validate();
		}

		showConnectionStatus("Connecting to " + host + ", port " + port + "...");

                mGazeProto = new GazeProto(host, port, this);
		rfb = mGazeProto;
		showConnectionStatus("Connected to server");

		rfb.readVersionMsg();
		showConnectionStatus("RFB server supports protocol version "
				+ rfb.serverMajor + "." + rfb.serverMinor);

		rfb.writeVersionMsg();
		showConnectionStatus("Using RFB protocol version " + rfb.clientMajor
				+ "." + rfb.clientMinor);

		int secType = rfb.negotiateSecurity();
		int authType;
		if (secType == RfbProto.SecTypeTight) {
			showConnectionStatus("Enabling TightVNC protocol extensions");
			rfb.setupTunneling();
			authType = rfb.negotiateAuthenticationTight();
		} else {
			authType = secType;
		}

		switch (authType) {
		case RfbProto.AuthNone:
			showConnectionStatus("No authentication needed");
			rfb.authenticateNone();
			break;
		case RfbProto.AuthVNC:
			showConnectionStatus("Performing standard VNC authentication");
			if (passwordParam != null) {
				rfb.authenticateVNC(passwordParam);
			} else {
				String pw = askPassword();
				rfb.authenticateVNC(pw);
			}
			break;
		default:
			throw new Exception("Unknown authentication scheme " + authType);
		}
    }

    @Override
    protected void createCanvas(int maxWidth, int maxHeight) throws IOException {
        // Determine if Java 2D API is available and use a special
        // version of VncCanvas if it is present.
        vc = null;
        // If we failed to create VncCanvas2D, use old VncCanvas.
        if (mGazeCanvas == null) {
            try {
                mGazeCanvas = new GazeVncCanvas(this, maxWidth, maxHeight);
            } catch (Exception ex) {
                Logger.getLogger(GazeVncViewer.class.getName()).log(Level.SEVERE, null, ex);
            }
        }

        vc = mGazeCanvas;
    }
}
